/** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* BiDiPro/AVR v0.1.4                                  avrc.c -- main module
* Copyright (c) Oleg O. Chukaev, 2007-2010  <oleg.chukaev@mail.ru>
* Based on AVRPP - AVR Parallel Programming Controller  R0.39 (C)ChaN, 2007
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **/


#pragma warn -par


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <malloc.h>
#include <errno.h>
#include <signal.h>

#include "defines.h"
#include "debug.h"
#include "hwctrl.h"
#include "avr.h"
#include "config.h"
#include "common.h"
#include "hex.h"
#include "cmdline.h"


#ifdef LINUXXX
int get_io_perm (void);
#endif


/* values for special_options */
#define SPECIAL_SHOW_HELP 0x01
#define SPECIAL_SHOW_VERSION 0x02
#define SPECIAL_SHOW_LICENSE 0x04


/* Forward declarations: cmd line parser handlers */
static int opt_special_handler (char *cmd, char *param, int extra);
static int opt_filename_handler (char *cmd, char *param, int extra);
static int opt_block_handler (char *cmd, char *param, int extra);
static int opt_ins_handler (char *cmd, char *param, int extra);
static int opt_debug_handler (char *cmd, char *param, int extra);
static int opt_copycal_handler (char *cmd, char *param, int extra);
static int opt_erase_handler (char *cmd, char *param, int extra);
static int opt_verify_handler (char *cmd, char *param, int extra);
static int opt_read_handler (char *cmd, char *param, int extra);
static int opt_fuse_handler (char *cmd, char *param, int extra);
static int opt_fusebyte_handler (char *cmd, char *param, int extra);
static int opt_lock_handler (char *cmd, char *param, int extra);
static int opt_voltage_handler (char *cmd, char *param, int extra);
static int opt_quick_handler (char *cmd, char *param, int extra);
static int opt_port_handler (char *cmd, char *param, int extra);
static int opt_devtype_handler (char *cmd, char *param, int extra);
static int opt_name_handler (char *cmd, char *param, int extra);

static void print_error (const char *fmt, ...);


/*-----------------------------------------------------------------------
  Global variables
-----------------------------------------------------------------------*/
static int cmd_read = 0;		/*  -, .  */
static int cmd_read_id = 0;		/* -ri  ID */
static int cmd_read_flash = 0;		/* -rp  Flash */
static int cmd_read_eeprom = 0;		/* -re  EEPROM */
static int cmd_read_fuses = 0;		/* -rf  Fuses */
static int cmd_erase = 0;		/* -e  Erase device */
static int cmd_write_fuses = 0;		/* -f, etc  Write fuses and/or lock */
static int cmd_verify = 0;		/* -v  Verify only */
static int cmd_copy_cal = 0;		/* -c  Copy calibration bytes into end of flash */
static int cmd_chips_help = 0;		/* -n help */
static int cmd_fuses_help = 0;		/* -f help */

static dword code_size = 0;		/* Loaded program code size (.hex) */
static dword data_size = 0;		/* Loaded EEPROM data size (.eep) */

static int ignore_errors = 0;		/* -I  Ignore non-fatal errors */
static int ins_flag = 0;		/* -i  Enable `insert chip` query */
static int fuses_default = 0;		/* Write def. fuses: 1,2,4,8: l,h,x,k */
static int special_options = 0;		/* -h, -L, -V */

static byte and_mask[4] = {0xff, 0xff, 0xff, 0xff};
static byte or_mask[4] = {0x00, 0x00, 0x00, 0x00};

static int file_p_mode = IO_HEX_STDOUT;	/*  []  Flash */
static int file_e_mode = IO_HEX_STDOUT;	/*  []  EEPROM */
//static int file_f_mode = 0;		/*  뢮 Fuses */

static char *forced_name = NULL;	/* -n Forced device name (compared to device->name) */
static char *fuses_text = NULL;		/* ப ࠬ  -f */
static char *file_p = NULL;		/*  䠩  ᮤন Flash */
static char *file_e = NULL;		/*  䠩  ᮤন EEPROM */
static char *file_f = NULL;		/*  䠩  ᮤন fuses */
static int first_port_mode = PAR;	/* 砫 ० ணࠬ஢ */
static int fuse_out_mode = 0;		/* 0: ⪨, 1: , 2:  */
char my_path[MAXPATH];			/*   avrc.exe */

/* command line options */
static opt_option options[] = {
	{'V', "version", 0, opt_special_handler, SPECIAL_SHOW_VERSION,
		"show version text", NULL},
	{'L', "license", 0, opt_special_handler, SPECIAL_SHOW_LICENSE,
		"show license text", NULL},
	{'h', "help", 0, opt_special_handler, SPECIAL_SHOW_HELP,
		"show help text", NULL},
	{'n', "name", 1, opt_name_handler, 0,
		"set chip name (`-n help' to list supported chips)", "name"},
	{'r', "read", 1, opt_read_handler, 0,
		"read ID / Flash / EEPROM / Fuses", "i|p|e|f[f[f]]"},
	{'v', "verify", 0, opt_verify_handler, 0,
		"verify Flash / EEPROM", NULL},
	{'e', "erase", 0, opt_erase_handler, 0,
		"erase Flash / EEPROM", NULL},
	{'f', "fuse", 1, opt_fuse_handler, 0,
		"write fuse bits (use `-f help' for help)", "values"},
	{'l', NULL, 0, opt_lock_handler, 0,
		"lock device (write both lock bits)", NULL},
	{0, "low", 1, opt_fusebyte_handler, 0,
		"write low fuse byte", "value"},
	{0, "high", 1, opt_fusebyte_handler, 1,
		"write high fuse byte", "value"},
	{0, "ext", 1, opt_fusebyte_handler, 2,
		"write extended fuse byte", "value"},
	{0, "lock", 1, opt_fusebyte_handler, 3,
		"write lock byte", "value"},
	{'c', "calib", 0, opt_copycal_handler, 0,
		"copy calibration value(s)", NULL},
	{'C', "code", 1, opt_filename_handler, 0,
		"name of file for Flash", "file"},
	{'D', "data", 1, opt_filename_handler, 1,
		"name of file for EEPROM", "file"},
	{'F', NULL, 1, opt_filename_handler, 2,
		"name of file for Fuses", "file"},
	{'q', "quick", 0, opt_quick_handler, 0,
		"quick power-up", NULL},
	{0, "vcc", 1, opt_voltage_handler, 1,
		"set Vcc", "voltage"},
	{0, "vpp", 1, opt_voltage_handler, 2,
		"set Vpp", "voltage"},
	{'p', "port", 1, opt_port_handler, 0,
		"set port number or port address", "port"},
	{'8', "hvs", 0, opt_devtype_handler, 1,
		"try HVS devices first", NULL},
	{'5', "t15", 0, opt_devtype_handler, 2,
		"try ATtiny15 first", NULL},
	{'0', "isp", 0, opt_devtype_handler, 3,
		"try ISP devices first", NULL},
	{'s', "block", 1, opt_block_handler, 0,
		"set HEX block size (for output)", "size"},
	{'i', NULL, 0, opt_ins_handler, 0,
		"enable `insert chip' query", NULL},
	{'d', "debug", 0, opt_debug_handler, 1,
		"enable debug output", NULL},
	{'Q', "quiet", 0, opt_debug_handler, 2,
		"disable debug and error output", NULL},
	{'I', "ignore", 0, opt_debug_handler, 3,
		"ignore non-fatal errors", NULL},
};


static const char hdr1[] =
	"avrc v0.1.4 - AVR programming tool for BiDiPro";
static const char hdr2[] =
	"Copyright (c) 2007-2010  Oleg O. Chukaev, <oleg.chukaev@mail.ru>";

/* version message */
static const char *version_msg[] = {
	"",
	hdr1,
	hdr2,
	"This program comes with ABSOLUTELY NO WARRANTY; for details type `avrc -L'.",
	"",
	"This software is based on:",
	"  - AVRPP - AVR Parallel Programming tool R0.39",
	"    (C)ChaN,2007  http://elm-chan.org/",
	"  - BiDiPro hardware driver v1.2",
	"    (C) Eduard Panchenko, Alex Plusov / CnC Lab, Andrej Arnold, 2003",
	"  - Command line parser for YASM",
	"    (C) 2001  Stanislav Karchebny <berk@madfire.net>",
	"",
	"See `avr.txt' or output of `avrc -h' for the list of options.",
	""
};

/* version message */
static const char *license_msg[] = {
	hdr1,
	hdr2,
	"",
	"This program is free software; you can redistribute it and/or modify",
	"it under the terms of the GNU General Public License as published by",
	"the Free Software Foundation; either version 2 of the License, or",
	"(at your option) any later version.",
	"",
	"This program is distributed in the hope that it will be useful,",
	"but WITHOUT ANY WARRANTY; without even the implied warranty of",
	"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the",
	"GNU General Public License for more details.",
	"",
	"You should have received a copy of the GNU General Public License",
	"along with this program; if not, write to the Free Software",
	"Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA",
	"",
};

/* help messages */
static const char *help_head =
	"usage: avrc [option]... [file]...\n"
	"Options:\n"
;

static const char *help_tail =
	"\n"
	"Sample invocation:\n"
	"   avrc -r p program.hex\n"
	"\n"
;


/****************************************************************************
****
****   -樨 뢮 ⫠筮  ᯮ⥫쭮 ଠ樨.
****
****************************************************************************/


/****************************************************************************
* 뢮  ⫠ ଠ.
* ࠬ:
*	---
* 頥:
*	---
****************************************************************************/

static void show_debug_info (void) {

	char *fn = "show_debug_info()";
	if (file_p) {
		debug_printf (2, "%s: Flash file: `%s'\n", fn, file_p);
		debug_printf (2, "%s: Flash file mode: `%d'\n", fn, file_p_mode);
	}
	if (file_e) {
		debug_printf (2, "%s: EEPROM file: `%s'\n", fn, file_e);
		debug_printf (2, "%s: EEPROM file mode: `%d'\n", fn, file_e_mode);
	}
	if (file_f)
		debug_printf (2, "%s: Fuses file: `%s'\n", fn, file_f);
	if (fuses_text)
		debug_printf (2, "%s: Fuses: `%s'\n", fn, fuses_text);
	if (forced_name)
		debug_printf (2, "%s: device name: `%s'\n", fn, forced_name);
	debug_printf (2, "%s: Vcc: %d, Vpp: %d\n", fn, vcc, vpp);
	debug_printf (2, "%s: port: %d\n", fn, port_number);
}


/****************************************************************************
* 뢮 ᮮ饭  ६ ᪠ .
* ࠬ:
*	---
* 頥:
*	---
****************************************************************************/

static void detect_print (void) {

	debug_printf (2, "%s mode: %s: %02X-%02X-%02X\n",
		port_modes[port_mode],
		(device != NULL) ? "device detected" : "unknown device",
		sign_buff[0], sign_buff[1], sign_buff[2]);
}


/****************************************************************************
* 뢠, ᪮쪮 業⮢ ⠫.
* ࠬ:
*	size -- 騩  
*	addr -- ⥪騩 
* 頥:
*	---
****************************************************************************/

static void show_progress (dword size, dword addr) {

	static long i = -1, j;

	j = addr * 100 / size;
	if (port_mode == PAR) {
		j = j / 10;
		j = j * 10;
	}
	if (i != j) {
		i = j;
		debug_printf (1, "%2ld%%\b\b\b", i);
	}
}


/****************************************************************************
* 뢮 ᮮ饭  訡  ஢થ .
* ࠬ:
*	addr --  訡
*	b1 --   
*	b2 --  ⠫
* 頥:
*	---
****************************************************************************/

static void show_failure (dword addr, byte b1, byte b2) {

	debug_printf (1, "failed at %04lX:%02X-%02X\n", addr, b1, b2);
}


/****************************************************************************
* 뢮 ᮮ饭  訡  . ப.
* ࠬ:
*	---
* 頥:
*	---
****************************************************************************/

static void print_error (const char *fmt, ...) {

	va_list va;

	va_start (va, fmt);
	vfprintf (stderr, fmt, va);
	va_end (va);
	fputc ('\n', stderr);
}


/****************************************************************************
*  Borland'᪮ strerror.   \n   ப.
****************************************************************************/

#ifndef LINUXXX
char *my_strerror (int errnum) {

	if (errnum < sys_nerr && errnum >= 0)
		return sys_errlist[errnum];
	else
		return "Unknown error :((";
}
#endif


/****************************************************************************
****
****   -樨 뢮 ଠ樨.
****
****************************************************************************/


/****************************************************************************
* 뢮 ᯨ᮪ ন ப஫஢.
* ࠬ:
*	---
* 頥:
*	---
****************************************************************************/

static void list_devices (void) {

	const DEVPROP *dev;
	int i;
	char *socket_names[] = {
		"---", "DIP-8", "DIP-14", "DIP-20 (T2313)",
		"DIP-20 (T26)", "DIP-28", "DIP-40 (M32)", "DIP-40 (M8515)",
	};

	printf ("Name            ID         Flash  EEPROM   Socket\n");
	printf ("-------------   --------   ------ ------   --------------\n");
	for (i = 0; (dev = dev_list[i]) != NULL; i++) {
		printf ("AT%-11s   %02X-%02X-%02X   %6ld   %4ld   %s\n", dev->name,
			dev->sign[0], dev->sign[1], dev->sign[2],
			dev->flash_size, dev->eeprom_size,
			socket_names[dev->socket]);
	}
}


/****************************************************************************
* 뢮 ଠ  .
* ࠬ:
*	---
* 頥:
*	---
* device 㪠뢠    ࠬࠬ 
****************************************************************************/

static void output_deviceinfo (void) {

	printf ("Device name: AT%s\n", device->name);
	printf ("Device signature: %02X-%02X-%02X\n",
		device->sign[0], device->sign[1], device->sign[2]);
	printf ("Flash memory size: %ld\n", device->flash_size);
	if (device->flash_page) {
		printf ("Flash memory page size: %d\n", device->flash_page);
		printf ("Number of flash pages: %d\n",
			(word)(device->flash_size / device->flash_page));
	}
	if (device->eeprom_size) {
		printf ("EEPROM size: %ld\n", device->eeprom_size);
		if (device->eeprom_page) {
			printf ("EEPROM page size: %d\n", device->eeprom_page);
			printf ("Number of EEPROM pages: %d\n",
				(word)(device->eeprom_size / device->eeprom_page));
		}
	}
}


/****************************************************************************
* 뢮 ଠ  fuse-  lock-.
* ࠬ:
*	num --   (0 .. 3  Low, High, Ext  Lock ᮮ.)
*	val -- 祭 
*	f -- 䠩, 㤠 뢮
* 頥:
*	---
****************************************************************************/

static void put_fuseval (int num, byte val, FILE *f) {

	int i, j, c;
#ifdef LINUXXX
	char line[] = "76543210-- %s";
#else
	char line[] = "76543210\xc4\xc4 %s";
#endif

	fprintf (f, "\n");
	for (i = 0; i < 8; i++) {
		if (fuse_names[num*8+7-i] == NULL) {
			c = '-';
		} else {
			c = (val & 0x80) ? '1' : '0';
		}
		fputc (c, f);
		val <<= 1;
	}
	switch (num) {
	case 0: fprintf (f, "   (Low)\n"); break;
	case 1: fprintf (f, "   (High)\n"); break;
	case 2: fprintf (f, "   (Ext)\n"); break;
	case 3: fprintf (f, "   (Lock)\n"); break;
	}
	for (i = 0; i < 8; i++) {
		if (fuse_names[num*8+i] != NULL) {
			for (j = 0; j < 8; j++) {
				if (j > i) {
#ifdef LINUXXX
					line[7-j] = (fuse_names[num*8+j] == NULL) ? ' ' : '|';
#else
					line[7-j] = (fuse_names[num*8+j] == NULL) ? ' ' : '\xb3';
#endif
				} else if (j == i) {
#ifdef LINUXXX
					line[7-j] = '`';
#else
					line[7-j] = '\xc0';
#endif
				} else {
#ifdef LINUXXX
					line[7-j] = '-';
#else
					line[7-j] = '\xc4';
#endif
				}
			}
			fprintf (f, line, fuse_names[num*8+i]);
			if (fuse_descr[num*8+i][0] != '\0') {
				fprintf (f, "  (%s)", fuse_descr[num*8+i]);
			}
			fprintf (f, "\n");
		}
	}
}


/****************************************************************************
* 뢮 ଠ  fuse-, lock-,  ஢ .
* ࠬ:
*	mode -- ० 뢮:
*		0 -- ॠ 祭
*		1 -- 䮫 祭
*	file --  䠩, 㤠 뢮, NULL --  stdout
* 頥:
*	0 --  
*	RC_FILE_WR -- ஡  䠩
* fuse_out_mode   ⢮ 뢮 ଠ樨
*  砥 뢮 䮫 祭 ᮤন ᨢ fuse_buff ࠧ蠥.
****************************************************************************/

/* FIXME -- ஢,  頥 fprintf? */
static int output_fuses (int mode, char *file) {

	int n, d;
	FILE *f;

	if (file) {
		f = fopen (file, "wt");
		if (f == NULL) {
			debug_printf (0, "error: can't open `%s': %s\n",
				file, strerror (errno));
			return (RC_FILE_WR);
		}
	} else {
		f = stdout;
	}

	if (mode == 1) {
		fuse_buff[0] = device->fuse_default[0];
		fuse_buff[1] = device->fuse_default[1];
		fuse_buff[2] = device->fuse_default[2];
		lock_byte = 0xff;
	}

	if (port_mode == ISP && device->fuse_type == 0) {
		debug_printf (0, "warning: fuse- and lock-bytes can't be read;"
			" using default values\n");
	}
	fprintf (f, "; Low: %02X", fuse_buff[0]);
	if (device->fuses >= 2)
		fprintf (f, "  High: %02X", fuse_buff[1]);
	if (device->fuses >= 3)
		fprintf (f, "  Ext: %02X", fuse_buff[2]);
	if (mode == 0) {
		fprintf (f, "  Lock: %02X", lock_byte);
		if (device->cals) {
			fprintf (f, "  Cal:");
			for (n = 0; n < device->cals; n++)
				fprintf (f, " %02X", cal_buff[n]);
		}
	}
	fprintf (f, "\n");

	if (fuse_out_mode == 0 || fuse_out_mode == 1) {
		for (n = 0; n < MAX_FUSES; n++) {
			if (fuse_names[n] != NULL) {
				if (n < 3 * 8) {
					d = (fuse_buff[n/8] >> n%8) & 1;
				} else {
					d = (lock_byte >> n%8) & 1;
				}
				fprintf (f, "%s=%d", fuse_names[n], d);
				if (fuse_out_mode == 1 && fuse_descr[n][0] != '\0') {
					fprintf (f, "  (%s)", fuse_descr[n]);
				}
				fprintf (f, "\n");
			}
		}
	} else if (fuse_out_mode == 2) {
		put_fuseval (0, fuse_buff[0], f);
		if (device->fuses >= 2) {
			put_fuseval (1, fuse_buff[1], f);
		}
		if (device->fuses >= 3) {
			put_fuseval (2, fuse_buff[2], f);
		}
		put_fuseval (3, lock_byte, f);
	}

	if (file) {
		if (fclose (f)) {
			return (RC_FILE_WR);
		}
	}
	return (RC_SUCCESS);
}


/****************************************************************************
* 樠 ணࠬ   .
* ࠬ:
*	power -- ⠢ ⠭ 祭  室  -樨  
*		0 --  ⠢
*		1 -- ⠢
* 頥:
*	0 -- Ok
*	RC_INIT_HW --  樠 ணࠬ
*	RC_INIT_DEV -- ஡  
*	RC_INIT_CFG -- ஡  䨣
****************************************************************************/

static int init_device (int power) {

	int rc;

	/* Execute initialization if not initialized yet */
	if (device != NULL)
		return (0);

#ifdef LINUXXX
	rc = get_io_perm ();
	debug_printf (2, "get_io_perm(): 0x%04x\n", rc);
	if (rc) {
		debug_printf (0, "error: can't get I/O permissions.\n");
		return (RC_INIT_HW);
	}
#endif

	rc = initialize_port ();
	debug_printf (2, "initialize_port(): 0x%04x\n", rc);
	if (rc) {
		debug_printf (0, "error: can't initialize hardware.\n");
		return (RC_INIT_HW);
	}

	if (ins_flag) {
		debug_printf (0, "Put a device on the socket and press Enter...");
		(void)getchar ();
	}

	if (forced_name) {
		rc = set_device (forced_name);
		if (rc) {
			debug_printf (0, "error: invalid device name: AT%s\n", forced_name);
			return (RC_INIT_DEV);
		} else {
			debug_printf (1, "Device is set to AT%s\n", forced_name);
		}
		if (port_mode == PAR) {
			if (strcmp (device->name, "tiny15") == 0) {
				first_port_mode = HVS15; port_mode = HVS15;
			} else if (device->socket == 0 || device->socket == 1) {
				first_port_mode = HVS; port_mode = HVS;
			}
		}
	} else {
		debug_printf (1, "Detecting a device. . .\n");
		rc = detect_device (first_port_mode, ISP, detect_print);
		switch (rc) {
		case DEVICE_FOUND:
			debug_printf (1, "Detected device is AT%s\n", device->name);
			break;
		case DEVICE_UNKNOWN:
			debug_printf (0, "error: no device is detected\n");
			return (RC_INIT_DEV);
		case DEVICE_LOCKED:
			debug_printf (0, "error: device is locked, use -n <device name> switch\n");
			return (RC_INIT_DEV);
		}
	}
	rc = fill_fuse_list ();
	if (rc) {
		debug_printf (0, "error: can't access `" FUSEFILE "'.  Use -d switch to see debug messages.\n");
		return (RC_INIT_CFG);
	}
	if (power)
		power_on ();
	return (RC_SUCCESS);
}


/****************************************************************************
****
****   -樨 ⥭ / ணࠬ஢.
****
****************************************************************************/


/****************************************************************************
* ࠥ .
* ࠬ:
*	---
* 頥:
*	0 -- Ok
*	RC_FAIL --   
****************************************************************************/

static int erase_device (void) {

	debug_printf (1, "Erasing. . . ");
	if (erase_memory ()) {
		debug_printf (1, "failed\n");
		return (RC_FAIL);
	}
	debug_printf (1, "Ok\n");

	return (0);
}


/****************************************************************************
* ⠥ 䨪, Flash, EEPROM, fuse-, lock-  ஢ .
* ࠬ:
*	---
* 頥:
*	0 -- Ok
*	 0 -- -  ⠪...
****************************************************************************/

static int read_device (void) {

	int rc = 0;

	/* -rp : read program memory */
	if (cmd_read_flash) {
		debug_printf (1, "Reading Flash. . . ");
		read_flash (show_progress);
		debug_printf (1, "Ok \n");
	}

	/* -re : read eeprom */
	if (cmd_read_eeprom) {
		if (device->eeprom_size == 0) {
			debug_printf (0, "warning: this device has no EEPROM\n");
		} else {
			debug_printf (1, "Reading EEPROM. . . ");
			read_eeprom (show_progress);
			debug_printf (1, "Ok \n");
		}
	}

	/* -rf : read fuses and cals */
	if (cmd_read_fuses) {
		read_fusecal ();
	}

	/* -ri */
	if (cmd_read_id) {
		output_deviceinfo ();
	}

	return (rc);
}


/****************************************************************************
*  / ஢ઠ Flash / EEPROM.
* ࠬ:
*	---
* 頥:
*	---
****************************************************************************/

static int write_device (void) {

	int n;
	char *f_str = ": flash: program size > memory size\n";
	char *e_str1 = ": device has no EEPROM\n";
	char *e_str2 = ": EEPROM: data size > memory size\n";

	/* Flash */
	if (code_size) {
		if (code_size > device->flash_size) {
			if (ignore_errors) {
				debug_printf (1, "warning%s", f_str);
			} else {
				debug_printf (0, "error%s", f_str);
				return (RC_FAIL);
			}
		}
		code_size = device->flash_size;

		/* -v : Skip programming process when verify mode */
		if (!cmd_verify) {
			/* Erase before writing */
			debug_printf (1, "Erasing. . . ");
			if (erase_memory ()) {
				debug_printf (1, "failed\n");
				return (RC_FAIL);
			}
			debug_printf (1, "Ok\n");

			/* -c : Copy calibration bytes */
			if (cmd_copy_cal && device->cals) {
				read_fusecal ();
				for (n = 0; n < device->cals; n++)
					code_buff[device->flash_size - 1 - n] = cal_buff[n];
				code_size = device->flash_size;
			}

			debug_printf (1, "Writing Flash. . . ");
			if (write_flash (show_progress)) {
				debug_printf (1, "failed\n");
				return (1);
			} else {
				debug_printf (1, "Ok \n");
			}
		}

		debug_printf (1, "Verifying Flash. . . ");
		if (verify_flash (show_progress, show_failure) == 0) {
			debug_printf (1, "Ok \n");
		}
	}

	/* EEPROM */
	if (data_size) {
		if (device->eeprom_size == 0) {
			if (ignore_errors) {
				debug_printf (1, "warning%s", e_str1);
			} else {
				debug_printf (0, "error%s", e_str1);
				return (RC_FAIL);
			}
		}

		if (data_size > device->eeprom_size) {
			if (ignore_errors) {
				debug_printf (1, "warning%s", e_str2);
			} else {
				debug_printf (0, "error%s", e_str2);
				return (RC_FAIL);
			}
		}

		/* -v : Skip programming process when verify mode */
		if (!cmd_verify) {
			debug_printf (1, "Writing EEPROM. . . ");
			if (write_eeprom (show_progress)) {
				debug_printf (1, "failed\n");
			} else {
				debug_printf (1, "Ok \n");
			}
		}

		debug_printf (1, "Verifying EEPROM. . . ");
		if (verify_eeprom (show_progress, show_failure) == 0) {
			debug_printf (1, "Ok \n");
		}
	}

	return (0);
}


/****************************************************************************
* 뢠 Fuse/Lock.
* ࠬ:
*	--
* 頥:
*	0 -- :)
*	RC_FAIL -- :(
****************************************************************************/

static int write_fuse (void) {

	byte fuse;

	read_fusecal ();

	if (cmd_write_fuses & 1) {
		fuse = (fuse_buff[0] & and_mask[0]) | or_mask[0] | ~device->fuse_mask[0];
		debug_printf (1, "writing fuse low byte (0x%02X). . . ", fuse);
		if (write_fuselock (F_LOW, fuse)) {
			debug_printf (1, "failed\n");
			return (RC_FAIL);
		}
		debug_printf (1, "Ok\n");
	}

	if (cmd_write_fuses & 2) {
		if (device->fuses >= 2) {
			fuse = (fuse_buff[1] & and_mask[1]) | or_mask[1]  | ~device->fuse_mask[1];
			debug_printf (1, "writing fuse high byte (0x%02X). . . ", fuse);
			if (write_fuselock (F_HIGH, fuse)) {
				debug_printf (1, "failed\n");
				return (RC_FAIL);
			}
			debug_printf (1, "Ok\n");
		} else {
			debug_printf (0, "warning: device has no high fuse byte\n");
		}
	}

	if (cmd_write_fuses & 4) {
		if (device->fuses >= 3) {
			fuse = (fuse_buff[2] & and_mask[2]) | or_mask[2] | ~device->fuse_mask[2];
			debug_printf (1, "writing fuse extended byte (0x%02X). . . ", fuse);
			if (write_fuselock (F_EXTEND, fuse)) {
				debug_printf (1, "failed\n");
				return (RC_FAIL);
			}
			debug_printf (1, "Ok\n");
		} else {
			debug_printf (0, "warning: device has no extended fuse byte\n");
		}
	}

	if (cmd_write_fuses & 8) {
		fuse = (lock_byte & and_mask[3]) | or_mask[3];
		debug_printf (1, "writing lock byte (0x%02X). . . ", fuse);
		if (write_fuselock (F_LOCK, fuse)) {
			debug_printf (1, "failed\n");
			return (RC_FAIL);
		}
		debug_printf (1, "Ok\n");
	}

	return (0);
}


/****************************************************************************
* ᢮ 뤥   ..
* ࠬ:
*	---
* 頥:
*	---
****************************************************************************/

static void terminate (void) {

	close_port ();
	device = NULL;
	kill_dev_list ();
	kill_fuse_list ();
	free (forced_name);
	free (fuses_text);
	free (file_p);
	free (file_e);
	free (file_f);
}


/****************************************************************************
* 㦠 䠩  .
* ࠬ:
*	---
* 頥:
*	0 -- Ok
*	RC_FILE_* -- :(
****************************************************************************/

static int load_files (void) {

	long ln;

	if (file_p != NULL) {
		ln = read_file (file_p_mode, file_p, code_buff, MAX_FLASH, &code_size);
		switch (ln) {
		case 0:
			debug_printf (2, "load_files(): Flash: read %ld bytes\n", code_size);
			break;
		case 1:
			debug_printf (0, "error: can't open `%s'\n", file_p);
			return (RC_FILE_RD);
		case 2:
		case 3:
			debug_printf (0, "error: can't access `%s'\n", file_p);
			return (RC_FILE_RD);
		case 4:
			debug_printf (0, "error: HEX format error in `%s'\n", file_p);
			return (RC_FILE_HEX);
		}
	}

	if (file_e != NULL) {
		ln = read_file (file_e_mode, file_e, data_buff, MAX_EEPROM, &data_size);
		switch (ln) {
		case 0:
			debug_printf (2, "load_files(): EEPROM: read %ld bytes\n", data_size);
			break;
		case 1:
			debug_printf (0, "error: can't open `%s'\n", file_e);
			return (RC_FILE_RD);
		case 2:
		case 3:
			debug_printf (0, "error: can't access `%s'\n", file_e);
			return (RC_FILE_RD);
		case 4:
			debug_printf (0, "error: HEX format error in `%s'\n", file_e);
			return (RC_FILE_HEX);
		}
	}

	if (file_f != NULL) {
		printf ("Fuse: Not yet supported :(\n");
	}
	return (RC_SUCCESS);
}


/****************************************************************************
* ࠭   䠩.
* ࠬ:
*	---
* 頥:
*	0 -- Ok
*	RC_FILE_* -- :(
****************************************************************************/

static int save_files (void) {

	int rc = 0;

	/* -rp : read program memory */
	if (cmd_read_flash) {
		debug_printf (2, "Saving Flash buffer. . .\n");
		rc = write_buffer (file_p_mode, file_p, code_buff, device->flash_size);
		if (rc) {
			debug_printf (0, "error: can't save `%s'\n", file_p);
			return (RC_FILE_WR);
		}
	} else if (file_p) {
		debug_printf (0, "warning: `%s' rejected\n", file_p);
	}

	/* -re : read eeprom */
	if (cmd_read_eeprom) {
		if (device->eeprom_size == 0) {
			debug_printf (0, "warning: this device has no EEPROM\n");
		} else {
			debug_printf (2, "Saving EEPROM buffer. . .\n");
			rc = write_buffer (file_e_mode, file_e, data_buff, device->eeprom_size);
			if (rc) {
				debug_printf (0, "error: can't save `%s'\n", file_e);
				return (RC_FILE_WR);
			}
		}
	} else if (file_e) {
		debug_printf (0, "warning: `%s' rejected\n", file_e);
	}

	/* -rf : read fuses and cals */
	if (cmd_read_fuses) {
		rc = output_fuses (0, file_f);
	} else if (file_f) {
		debug_printf (0, "warning: `%s' rejected\n", file_f);
	}

	return (rc);
}


/****************************************************************************
* 뤥  ப  fuse1=0,fuse2=1,...   祭 fuse-⮢.
* ࠬ:
*	dst -- 㤠 ஢ 
*	src -- ப
*	c -- 祭 fuse-
* 頥:
*	0 --  ப
*	> 0 --  ,  ண த
*	< 0 -- 訡
****************************************************************************/

static int get_fuse (char *dst, char *src, char *c) {

	int i = 0, j = 0;
	char cc = '-';

	if (src[i] == ',' || src[i] == '=') {
		return (-1);
	}
	while (src[i] != ',' && src[i] != '\0') {
		if (src[i] != '=') {
			dst[j] = src[i];
			i++; j++;
		} else {
			i++;
			if (src[i] != '0' && src[i] != '1' && src[i] != '*')
				return (-2);
			cc = src[i++];
			dst[j] = '\0';
			break;
		}
	}
	if (cc == '-' && src[0] != '\0')
		return (-3);
	*c = cc;
	if (src[i] == ',')
		i++;
	return (i);
}


/****************************************************************************
* ࠥ ப  fuse-⠬.
* ࠬ:
*	---
* 頥:
*	0 -- Ok
*	RC_SYNTAX -- ᨭ⠪᪠ 訡
****************************************************************************/

static int parse_fuses (char *ft) {

	int i, j, k, found;
	char c, str[150];

	i = j = 0;
	while ((i = get_fuse (str, ft+j, &c)) > 0) {
		j += i;
		found = 0;
		for (k = 0; k < MAX_FUSES; k++) {
			if (fuse_names[k] != NULL &&
				stricmp (fuse_names[k], str) == 0)
			{
				found = 1;
				if (c == '*' && k < 3 * 8) {
					c = device->fuse_default[k/8] &
						(1 << k%8) ? '1' : '0';
				} else if (c == '*' && k >= 3 * 8) {
					c = device->lock_data &
						(1 << k%8) ? '1' : '0';
				}
				if (c == '0') {
					and_mask[k/8] &= ~(1 << k%8);
					or_mask[k/8] &= ~(1 << k%8);
				} else if (c == '1') {
					and_mask[k/8] |= (1 << k%8);
					or_mask[k/8] |= (1 << k%8);
				}
				cmd_write_fuses |= 1 << k/8;
				debug_printf (2, "parse_fuses(): %s=%c\n", fuse_names[k], c);
				break;
			}
		}
		if (!found) {
			debug_printf (0, "error: unknown fuse: `%s'\n", str);
			return (RC_SYNTAX);
		}
	}
	if (i < 0) {
		debug_printf (0, "error: invalid parameter: `%s'\n", ft+j);
		return (RC_SYNTAX);
	}

	return (RC_SUCCESS);
}


/****************************************************************************
* ࠥ 䠩  fuse'.
* ࠬ:
*	---
* 頥:
*	0 -- Ok
*	RC_FILE_RD -- 訡 /⥭ 䠩
*	RC_SYNTAX -- ᨭ⠪᪠ 訡
****************************************************************************/

static int parse_fuses_file (void) {

	int linenum, rc;
	char line[200];
	char *z;
	FILE *f;

	f = fopen (file_f, "rt");
	if (f == NULL) {
		debug_printf (0, "error: can't open `%s': %s\n",
			file_f, strerror (errno));
		return (RC_FILE_RD);
	}

	linenum = 0;
	while (fgets (line, sizeof(line), f) != NULL) {
		linenum++;
		z = strchr (line, '\r'); if (z) *z = '\0';
		z = strchr (line, '\n'); if (z) *z = '\0';
		if (line[0] != '#' && line[0] != ';' && line[0] != '\0') {
			rc = parse_fuses (line);
			if (rc) {
				debug_printf (0, "parse_fuses_file(): %s: error"
					" at line: %d\n", file_f, linenum);
				fclose (f);
				return (rc);
			}
		}
	}

	fclose (f);
	return (RC_SUCCESS);
}


/****************************************************************************
* ࠥ 䮫 祭 fuse-⮢.
* ࠬ:
*	---
* 頥:
*	---
****************************************************************************/

static void parse_fuses_default (void) {

	char *fn = "parse_fuses_default()";

	if (fuses_default & 1) {
		and_mask[0] = or_mask[0] = device->fuse_default[0];
		debug_printf (2, "%s: low=%02X\n", fn, or_mask[0]);
	}
	if (fuses_default & 2) {
		and_mask[1] = or_mask[1] = device->fuse_default[1];
		debug_printf (2, "%s: high=%02X\n", fn, or_mask[1]);
	}
	if (fuses_default & 4) {
		and_mask[2] = or_mask[2] = device->fuse_default[2];
		debug_printf (2, "%s: ext=%02X\n", fn, or_mask[2]);
	}
	if (fuses_default & 8) {
		and_mask[3] = or_mask[3] = device->lock_data;
		debug_printf (2, "%s: lock=%02X\n", fn, or_mask[3]);
	}
}


/****************************************************************************
****
****   -樨  ࠧ  ப.
****
****************************************************************************/

/*  䠩 */
int not_an_option_handler (char *param) {

	char s[200];
	char str[] = "warning: `%s' rejected: you should specify only one"
		" file for %s\n";
	char str2[] = "error: bad filename: `%s'\n";

	strncpy (s, param, sizeof(s)); s[sizeof(s)-1] = '\0';
#ifndef LINUXXX
	strlwr (s);
#endif
	if (strlen (s) < 5) {
		debug_printf (0, str2, param);
		return (1);
	}

	if (strcmp (s+strlen(s)-4, ".hex") == 0) {
		if (file_p) {
			debug_printf (0, str, param, "Flash");
		} else {
			file_p = strdup (param);
			file_p_mode = IO_HEX_FILE;
		}
		return (0);
	} else if (strcmp (s+strlen(s)-4, ".eep") == 0) {
		if (file_e) {
			debug_printf (0, str, param, "EEPROM");
		} else {
			file_e = strdup (param);
			file_e_mode = IO_HEX_FILE;
		}
		return (0);
	} else if (strcmp (s+strlen(s)-4, ".fus") == 0) {
		if (file_f) {
			debug_printf (0, str, param, "Fuses");
		} else {
			file_f = strdup (param);
		}
		return (0);
	}
	debug_printf (0, str2, param);
	return (1);
}


/*   */
int other_option_handler (char *option) {

	return 1;
}


/* -h, -L, -V */
static int opt_special_handler (char *cmd, char *param, int extra) {

	if (special_options == 0)
		special_options = extra;
	return (0);
}


/* -C, -D, -F */
static int opt_filename_handler (char *cmd, char *param, int extra) {

	char s[200];
	char str[] = "warning: `%s' rejected: you should specify only one"
		" file for %s\n";

	strncpy (s, param, sizeof(s)); s[sizeof(s)-1] = '\0';
#ifndef LINUXXX
	strlwr (s);
#endif
	switch (extra) {
	case 0:
		if (file_p) {
			debug_printf (0, str, param, "Flash");
		} else {
			if ((strlen(s) > 4) &&
				(strcmp (s+strlen(s)-4, ".hex") == 0))
			{
				file_p_mode = IO_HEX_FILE;
			} else {
				file_p_mode = IO_BIN_FILE;
			}
			file_p = strdup (param);
		}
		break;
	case 1:
		if (file_e) {
			debug_printf (0, str, param, "EEPROM");
		} else {
			if ((strlen(s) > 4) &&
				((strcmp (s+strlen(s)-4, ".hex") == 0)
				|| (strcmp (s+strlen(s)-4, ".eep") == 0)))
			{
				file_e_mode = IO_HEX_FILE;
			} else {
				file_e_mode = IO_BIN_FILE;
			}
			file_e = strdup (param);
		}
		break;
	case 2:
		if (file_f) {
			debug_printf (0, str, param, "Fuses");
		} else {
			file_f = strdup (param);
		}
		break;
	}
	return (0);
}


/* -s */
static int opt_block_handler (char *cmd, char *param, int extra) {

	hex_block_size = (int)atol (param);
	return (0);
}


/* -i */
static int opt_ins_handler (char *cmd, char *param, int extra) {

	ins_flag = 1;
	return (0);
}


/* -d, -Q, -I */
static int opt_debug_handler (char *cmd, char *param, int extra) {

	switch (extra) {
	case 1:
		/* -d */
		debug_level++; break;
	case 2:
		/* -Q */
		debug_level = -1; break;
	case 3:
		/* -I */
		ignore_errors = 1; break;
	}
	return (0);
}


/* -c */
static int opt_copycal_handler(char *cmd, char *param, int extra) {

	cmd_copy_cal = 1;
	return (0);
}


/* -e */
static int opt_erase_handler (char *cmd, char *param, int extra) {

	cmd_erase = 1;
	return (0);
}


/* -v */
static int opt_verify_handler (char *cmd, char *param, int extra) {

	cmd_verify = 1;
	return (0);
}


/* -r */
static int opt_read_handler (char *cmd, char *param, int extra) {

	if (strcmp (param, "i") == 0) {
		cmd_read_id = 1;
	} else if (strcmp (param, "p") == 0) {
		cmd_read_flash = 1;
	} else if (strcmp (param, "e") == 0) {
		cmd_read_eeprom = 1;
	} else if (strcmp (param, "f") == 0) {
		cmd_read_fuses = 1;
	} else if (strcmp (param, "ff") == 0) {
		cmd_read_fuses = 1;
		fuse_out_mode = 1;
	} else if (strcmp (param, "fff") == 0) {
		cmd_read_fuses = 1;
		fuse_out_mode = 2;
	} else {
		return (1);
	}
	cmd_read = 1;
	return (0);
}


/* -f, --fuse */
static int opt_fuse_handler (char *cmd, char *param, int extra) {

	if (strcmp (param, "help") == 0) {
		cmd_fuses_help = 1;
		fuse_out_mode = 0;
	} else if (strcmp (param, "heelp") == 0) {
		cmd_fuses_help = 1;
		fuse_out_mode = 1;
	} else if (strcmp (param, "heeelp") == 0) {
		cmd_fuses_help = 1;
		fuse_out_mode = 2;
	} else {
		if (fuses_text)
			free (fuses_text);
		fuses_text = strdup (param);
	}
	return 0;
}


/* --low, --high, --ext, --lock */
static int opt_fusebyte_handler (char *cmd, char *param, int extra) {

	long i;
	char *p;

	if (strcmp (param, "*") == 0) {
		cmd_write_fuses |= 1 << extra;
		fuses_default |= 1 << extra;
		return (0);
	}
	if (strncmp (param, "0x", 2) == 0) {
		i = strtol (param+2, &p, 16);
	} else {
		i = strtol (param, &p, 2);
	}
	if (p < param + strlen (param) || i > 255)
		return (1);
	and_mask[extra] = i;
	or_mask[extra] = i;
	cmd_write_fuses |= 1 << extra;
	return (0);
}


/* -l */
static int opt_lock_handler (char *cmd, char *param, int extra) {

	fuses_default |= 1 << 3;	/* Same as --lock=* */
	cmd_write_fuses |= 1 << 3;
	return (0);
}


/* --vcc, --vpp */
static int opt_voltage_handler (char *cmd, char *param, int extra) {

	long i;

	i = atol (param);
	if (i == 0)
		return (1);

	switch (extra) {
	case 1:
		/* --vcc */
		vcc = i; break;
	case 2:
		/* --vpp */
		vpp = i; break;
	}
	return (0);
}


/* -q */
static int opt_quick_handler (char *cmd, char *param, int extra) {

	quick_power = 1;
	return (0);
}


/* -p */
static int opt_port_handler (char *cmd, char *param, int extra) {

	char *p;
	long i;

	if (strncmp (param, "0x", 2) == 0) {
		i = strtol (param+2, &p, 16);
	} else {
		i = strtol (param, &p, 10);
	}
	if (p < param + strlen (param) || i < 0)
		return (1);
	port_number = (int)i;
	return (0);
}


/* -8, -5, -0 */
static int opt_devtype_handler (char *cmd, char *param, int extra) {

	switch (extra) {
	case 1:
		/* -8 */
		first_port_mode = HVS; port_mode = HVS; break;
	case 2:
		/* -5 */
		first_port_mode = HVS15; port_mode = HVS15; break;
	case 3:
		/* -0 */
		first_port_mode = ISP; port_mode = ISP; break;
	}
	return (0);
}


/* -n */
static int opt_name_handler (char *cmd, char *param, int extra) {

	if (strcmp (param, "help") == 0) {
		cmd_chips_help = 1;
	} else {
		if (strnicmp (param, "at", 2) == 0)
			param += 2;
		forced_name = strdup (param);
	}
	return (0);
}


/****************************************************************************
* ࠡ稪 ᨣ SIGTERM, SIGINT, SIGHUP, SIGQUIT.
* ࠬ:
*	sig --  ᨣ
* 頥:
*	---
****************************************************************************/

static void signal_handler (int sig) {

	printf ("\nSignal %d caught, exiting...\n", sig);
	terminate ();
	exit (RC_INT);
}


/****************************************************************************
****
****   main ()
****
****************************************************************************/

int main (int argc, char **argv) {

	int rc, cmd;
	char *p, *z;
	size_t i;
	char *cp, *cmdlst[40], cmdbuff[1000], line[200];
	FILE *fp;

	strncpy (my_path, argv[0], sizeof(my_path));
	my_path[sizeof(my_path)-1] = '\0';
	p = strrchr (my_path, PATH_SEPARATOR);
	*++p = '\0';
	strncpy (progname, "avrc", PROGNAME_LEN);
	progname[PROGNAME_LEN-1] = '\0';

	signal (SIGINT, signal_handler);
	signal (SIGTERM, signal_handler);
#ifdef LINUXXX
	signal (SIGHUP, signal_handler);
	signal (SIGQUIT, signal_handler);
#endif

	code_buff = farmalloc (MAX_FLASH);
	if (code_buff == NULL) {
		debug_printf (0, "error: can't allocate memory for buffers\n");
		return (RC_INIT_MEM);
	}
	/* Clear data buffers */
	memset (code_buff, 0xff, MAX_FLASH);
	memset (data_buff, 0xff, MAX_EEPROM);

	/* Import ini file as command line parameters */
	cmd = 1; cp = cmdbuff+1; cmdbuff[0] = '\0'; cmdlst[0] = cmdbuff;
	fp = open_cfgfile (INIFILE);
	if (fp != NULL) {
		while (fgets (line, sizeof(line), fp) != NULL) {
			z = strchr (line, '\r');
			if (z) *z = '\n';
			if (line[0] != ';' && line[0] != '\n') {
				if (cmd >= NELEMS(cmdlst) - 1)
					break;
				strncpy (cp, line, cmdbuff + sizeof(cmdbuff) - cp);
				cmdlst[cmd++] = cp;
				z = strchr (cp, '\n');
				if (z) *z = '\0';
				cp += strlen (cp) + 1;
			}
		}
		fclose (fp);
	}

	if (parse_cmdline (cmd, cmdlst, options, NELEMS(options), print_error)) {
		return (RC_SYNTAX);
	}
	if (parse_cmdline (argc, argv, options, NELEMS(options), print_error)) {
		return (RC_SYNTAX);
	}

	show_debug_info ();

	switch (special_options) {
	case SPECIAL_SHOW_HELP:
		/* -h, --help */
		help_msg (help_head, help_tail, options, NELEMS(options));
		return (RC_SUCCESS);
	case SPECIAL_SHOW_VERSION:
		/* -V, --version */
		for (i = 0; i < NELEMS(version_msg); i++)
			printf ("%s\n", version_msg[i]);
		return (RC_SUCCESS);
	case SPECIAL_SHOW_LICENSE:
		/* -L, --license */
		for (i = 0; i < NELEMS(license_msg); i++)
			printf("%s\n", license_msg[i]);
		return (RC_SUCCESS);
	}

	rc = fill_dev_list ();
	if (rc) {
		debug_printf (0, "error: can't access `" CHIPSFILE "'.  Use -d switch(es) to see debug messages.\n");
		return (RC_INIT_CFG);
	}

	/* -n help */
	if (cmd_chips_help) {
		list_devices ();
		kill_dev_list ();
		return (RC_SUCCESS);
	}

	/* -f help */
	if (cmd_fuses_help) {
		if (forced_name) {
			rc = set_device (forced_name);
			if (rc) {
				debug_printf (0, "error: invalid device name: AT%s\n", forced_name);
				return (RC_INIT_DEV);
			} else {
				debug_printf (1, "Device is set to AT%s\n", forced_name);
			}
			rc = fill_fuse_list ();
			if (rc) {
				debug_printf (0, "error: can't access `" FUSEFILE "'.  Use -d switch to see debug messages.\n");
				return (RC_INIT_CFG);
			}
		} else {
			rc = init_device (0);
			if (rc)
				return (rc);
		}
		rc = output_fuses (1, file_f);
		if (rc)
			return (rc);
		return (RC_SUCCESS);
	}

	/* -ri, -rp, -re, -rf, --read */
	if (cmd_read) {
		rc = init_device (1);
		if (rc) {
			terminate ();
			return (rc);
		}
		rc = read_device ();
		if (rc) {
			terminate ();
			return (RC_READ);
		}
		rc = save_files ();
		terminate ();
		if (rc)
			return (rc);
		return (RC_SUCCESS);
	}

	/* -e, --erase */
	if (cmd_erase) {
		rc = init_device (1);
		if (rc) {
			terminate ();
			return (rc);
		}
		rc = erase_device ();
		terminate ();
		if (rc)
			return (RC_ERASE);
		return (RC_SUCCESS);
	}

	if (file_f) {
		rc = init_device (1);
		if (rc) {
			terminate ();
			return (rc);
		}
		rc = parse_fuses_file ();
		if (rc) {
			terminate ();
			return (rc);
		}
	}
	if (fuses_default) {
		rc = init_device (1);
		if (rc) {
			terminate ();
			return (rc);
		}
		parse_fuses_default ();
	}
	if (fuses_text) {
		rc = init_device (1);
		if (rc) {
			terminate ();
			return (rc);
		}
		rc = parse_fuses (fuses_text);
		if (rc) {
			terminate ();
			return (rc);
		}
	}

	if (file_p || file_e || cmd_write_fuses) {
		/* Write Flash/EEPROM */
		if (file_p || file_e) {
			rc = load_files ();
			if (rc) {
				terminate ();
				return (rc);
			}
			rc = init_device (1);
			if (rc) {
				terminate ();
				return (rc);
			}
			rc = write_device ();
			if (rc) {
				terminate ();
				return (RC_WRITE);
			}
		}

		/* -f, --fuse, -l, --low, --high, --ext, --lock */
		if (cmd_write_fuses) {
			rc = init_device (1);
			if (rc) {
				terminate ();
				return (rc);
			}
			rc = write_fuse ();
			if (rc) {
				terminate ();
				return (RC_FUSES);
			}
		}

		terminate ();
		return (RC_SUCCESS);
	}

	terminate ();

	/* 祣 ⢥ */
	for (i = 0; i < NELEMS(version_msg); i++)
		printf ("%s\n", version_msg[i]);

	return (RC_SUCCESS);
}


/** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* E0F
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **/

