/** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* BiDiPro/AVR v0.1.4                  common.c -- common functions and data
* Copyright (c) Oleg O. Chukaev, 2008-2010  <oleg.chukaev@mail.ru>
* Based on AVRPP - AVR Parallel Programming Controller  R0.39 (C)ChaN, 2007
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **/


#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include "defines.h"
#include "avr.h"
#include "hwctrl.h"
#include "hex.h"
#include "debug.h"

byte *code_buff;		/* Program code R/W buffer */
byte data_buff[MAX_EEPROM];	/* EEPROM data R/W buffer */

byte cal_buff[4];		/* Calibration bytes read buffer */
byte fuse_buff[3];		/* Fuse bytes read buffer */
byte sign_buff[3];		/* Device signature read buffer */
byte lock_byte;			/* Lock byte */

DEVPROP *dev_list[MAX_DEVICES];	/* ᨢ 㪠⥫  .  */
const DEVPROP *device = NULL;	/* ⥫  ࠬ ⥪饣  */

const char *const port_modes[] = {"PAR", "HVS", "HVS15", "HVS15_OLD", "ISP"};

char *fuse_names[MAX_FUSES];	/* ᨢ 㪠⥫   fuse-⮢ */
char *fuse_descr[MAX_FUSES];	/* ᨢ 㪠⥫  ᠭ fuse-⮢ */



/****************************************************************************
*    ⠡    (ॣ 㪢 )
* ࠬ:
*	name --  
* 頥:
*	0 --  , device 㪠뢠   
*	1 --   , device == NULL
****************************************************************************/

int set_device (char *name) {

	int i;

	for (i = 0; (device = dev_list[i]) != NULL; i++) {
		/* !!! FIXME !!!  (   stricmp) */
		if (stricmp (name, device->name) == 0)
			return (0);
	}
	return (1);
}


/****************************************************************************
* ⮬᪨ । .
* ࠬ:
*	f_mode --  ࢮ ०
*	l_mode --  ᫥ ०
*	print -- - 뢮 ᮮ饭, NULL, ᫨  ᯮ
* 頥:
*	DEVICE_FOUND --  , device 㪠뢠   ,
*		sign_buff ᮤন 䨪 , port_mode
*		ᮤন  ⥪饣 ०
*	DEVICE_UNKNOWN --   
*	DEVICE_LOCKED --  祭
* ᫨ 㭪樨 । ⨬ 祭 ࠬ஢ f_mode /
* l_mode,  f_mode := PAR, l_mode := ISP
****************************************************************************/

int detect_device (int f_mode, int l_mode, void (*print)(void)) {

	dword adr;
	int i;

	if ((f_mode < PAR) || (l_mode > ISP) || (f_mode > l_mode)) {
		f_mode = PAR;
		l_mode = ISP;
	}

	for (port_mode = f_mode; port_mode <= l_mode; port_mode++) {
		power_on ();

		/* read device signature */
		for (adr = 0; adr < 3; adr++)
			sign_buff[adr] = read_byte (SIGNATURE, adr);

		/*  祭? */
		if ((port_mode == ISP) && (sign_buff[0] == 0)
			&& (sign_buff[1] == 1) && (sign_buff[2] == 2))
		{
			power_off ();
			delay_ms (50);
			return (DEVICE_LOCKED);
		}

		/* search device table */
		for (i = 0; (device = dev_list[i]) != NULL; i++) {
			if (memcmp (sign_buff, device->sign, 3) == 0)
				break;
		}

 		power_off ();

		if (print != NULL)
			print ();

 		if (device != NULL)
			return (DEVICE_FOUND);

		delay_ms (50);
 	}
	return (DEVICE_UNKNOWN);
}


/****************************************************************************
*    䠩 / stdout.
* ࠬ:
*	dst -- IO_HEX_STDOUT, IO_HEX_FILE, IO_BIN_FILE
*	fname --  䠩
*	addr --  
*	size -- ࠧ 
* 頥:
*	0 -- Ok
*	1 --  
*	2 --  
*	3 -- ⨬ 祭 dst
****************************************************************************/

int write_buffer (int dst, char *fname, byte *addr, dword size) {

	FILE *fil = NULL;
	int rc, items;
	char my_name[] = "write_buffer()";

	switch (dst) {
	case IO_HEX_STDOUT:
		rc = output_hexfile (stdout, addr, size, 0);
		if (rc)
			return (2);
		break;

	case IO_HEX_FILE:
		fil = fopen (fname, "wt");
		if (fil == NULL) {
			debug_printf (1, "%s: can't open `%s': %s\n",
				my_name, fname, strerror (errno));
			return (1);
		}
		rc = output_hexfile (fil, addr, size, 0);
		if (rc) {
			debug_printf (1, "%s: can't write `%s': output_hexfile() failed\n",
				my_name, fname);
			fclose (fil);
			return (2);
		}
		break;

	case IO_BIN_FILE:
		fil = fopen (fname, "wb");
		if (fil == NULL) {
			debug_printf (1, "%s: can't open `%s': %s\n",
				my_name, fname, strerror (errno));
			return (1);
		}
		if (size <= 32768) {
			items = size;
			rc = fwrite (addr, 1, items, fil);
		} else {
			items = size / 512L;
			rc = fwrite (addr, 512, items, fil);
		}
		if (rc < items) {
			debug_printf (1, "%s: can't write `%s': %s\n",
				my_name, fname, strerror (errno));
			return (2);
		}
		rc = fclose (fil);
		if (rc) {
			debug_printf (1, "%s: can't close `%s': %s\n",
				my_name, fname, strerror (errno));
			return (2);	/* FIXME? */
		}
		break;

	default:
		return (3);
	}

	return (0);
}


/****************************************************************************
* ⥭ 䠩  .
* ࠬ:
*	type -- IO_HEX_FILE, IO_BIN_FILE
*	fname --  䠩
*	addr --  
*	size -- ࠧ 
*	datasize -- ᪮쪮  ॠ쭮 ⠭
* 頥:
*	0 -- Ok
*	1 --  
*	2 --  
*	3 -- ⨬ 祭 type
*	4 -- 訡  HEX-䠩
****************************************************************************/

int read_file (int type, char *fname, byte *addr, dword size, dword *datasize) {

	FILE *fil;
	long rc, rdb, ptr;
	char my_name[] = "read_file()";

	switch (type) {
	case IO_HEX_FILE:
		fil = fopen (fname, "rt");
		if (fil == NULL) {
			debug_printf (1, "%s: can't open `%s': %s\n",
				my_name, fname, strerror (errno));
			return (1);
		}
		rc = input_hexfile (fil, addr, size, datasize);
		if (rc == -1) {
			debug_printf (1, "%s: can't read `%s': input_hexfile() failed\n",
				my_name, fname);
			fclose (fil);
			return (2);
		} else if (rc > 0) {
			debug_printf (1, "%s: can't read `%s': error at line %ld\n",
				my_name, fname, rc);
			fclose (fil);
			return (4);
		}
		fclose (fil);
		break;

	case IO_BIN_FILE:
		fil = fopen (fname, "rb");
		if (fil == NULL) {
			debug_printf (1, "%s: can't open `%s': %s\n",
				my_name, fname, strerror (errno));
			return (1);
		}
		ptr = 0L;
		do {
			if (size - ptr < 32768)
				rdb = size - ptr;
			else
				rdb = 32768;
			rc = fread (addr+ptr, 1, rdb, fil);
			ptr += rc;
		} while (ptr < size && !feof (fil) && !ferror (fil));
		if (ferror (fil)) {
			debug_printf (1, "%s: can't read `%s': fread() failed\n",
				my_name, fname);
			fclose (fil);
			return (2);
		}
		*datasize = ptr;
		fclose (fil);
		break;

	default:
		return (3);
	}

	return (0);
}


/****************************************************************************
* ⠥ Flash / EEPROM  ᮮ⢥騩 .
* ࠬ:
*	print(size, addr) -- - 뢮 ⥪饣 ﭨ, NULL, ᫨ 
*	ᯮ.
*		size -- ࠧ 
*		addr -- ⥪騩 
* 頥:
*	0 -- Ok
****************************************************************************/

int read_flash (void (*print)(dword, dword)) {

	dword adr;

	for (adr = 0; adr < device->flash_size; adr++) {
		code_buff[adr] = read_byte (FLASH, adr);
		if (print != NULL) {
			print (device->flash_size, adr);
		}
	}
	return (0);
}


int read_eeprom (void (*print)(dword, dword)) {

	dword adr;

	for (adr = 0; adr < device->eeprom_size; adr++) {
		data_buff[adr] = read_byte (EEPROM, adr);
		if (print != NULL) {
			print (device->eeprom_size, adr);
		}
	}
	return (0);
}


/****************************************************************************
* ⠥ Fuse-, Lock-  ஢   ᮮ. .
* ࠬ:
*	---
* 頥:
*	---
****************************************************************************/

void read_fusecal (void) {

	dword adr = 0;

	do {
		fuse_buff[adr] = read_byte (FUSE, adr);
	} while ((char)(++adr) < device->fuses);

	for (adr = 0; (char)adr < device->cals; adr++)
		cal_buff[adr] = read_byte (CALIBS, adr);

	lock_byte = read_byte (FUSE, 3);
}


/****************************************************************************
* 뢠 ᮮ.   Flash / EEPROM.
* ࠬ:
*	---
* 頥:
*	0 -- Ok
*	1 -- 訡 
****************************************************************************/

int write_flash (void (*print)(dword, dword)) {

	dword adr;

	if (device->flash_page) {
		/* Write flash in page mode */
		for (adr = 0; adr < device->flash_size; adr += device->flash_page) {
			if (print != NULL) {
				print (device->flash_size, adr);
			}
			if (write_page (FLASH, adr, &code_buff[adr])) {
				return (1);
			}
		}
	} else {
		/* Write flash in byte-by-byte mode */
		for (adr = 0; adr < device->flash_size; adr++) {
			if (print != NULL) {
				print (device->flash_size, adr);
			}
			if (write_byte (FLASH, adr, code_buff[adr])) {
				return (1);
			}
		}
	}
	return (0);
}


int write_eeprom (void (*print)(dword, dword)) {

	dword adr;

	if (device->eeprom_page && port_mode != ISP) {
		/* Write eeprom in page mode */
		for (adr = 0; adr < device->eeprom_size; adr += device->eeprom_page) {
			if (print != NULL) {
				print (device->eeprom_size, adr);
			}
			if (write_page (EEPROM, adr, &data_buff[adr])) {
				return (1);
			}
		}
	} else {
		/* Write eeprom in byte-by-byte mode */
		for (adr = 0; adr < device->eeprom_size; adr++) {
			if (print != NULL) {
				print (device->eeprom_size, adr);
			}
			if (write_byte (EEPROM, adr, data_buff[adr])) {
				return (1);
			}
		}
	}
	return (0);
}


/****************************************************************************
* ࠢ Flash / EEPROM  ஬.
* ࠬ:
*	print(size, addr) -- - 뢮 ⥪饣 ﭨ, NULL, ᫨ 
*	ᯮ.
*		size -- ࠧ 
*		addr -- ⥪騩 
*	fail(addr, buf, rd) -- - 뢮 ᮮ饭  ᮢ, NULL,
*	᫨  ᯮ.
*		addr --  ᮢ
*		buf --   
*		rd --   
* 頥:
*	0 -- Ok
*	1 -- 訡 ࠢ
****************************************************************************/

int verify_flash (void (*print)(dword, dword), void (*fail)(dword, byte, byte)) {

	dword adr;
	byte rd;

	for (adr = 0; adr < device->flash_size; adr++) {
		rd = read_byte (FLASH, adr);
		if (print != NULL) {
			print (device->flash_size, adr);
		}
		if (rd != code_buff[adr]) {
			if (fail != NULL) {
				fail (adr, code_buff[adr], rd);
			}
			return (1);
		}
	}
	return (0);
}


int verify_eeprom (void (*print)(dword, dword), void (*fail)(dword, byte, byte)) {

	dword adr;
	byte rd;

	for (adr = 0; adr < device->eeprom_size; adr++) {
		rd = read_byte (EEPROM, adr);
		if (print != NULL) {
			print (device->eeprom_size, adr);
		}
		if (rd != data_buff[adr]) {
			if (fail != NULL) {
				fail (adr, data_buff[adr], rd);
			}
			return (1);
		}
	}
	return (0);
}


/** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* E0F
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **/
