/** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* BiDiPro/AVR v0.1.4                                               config.c
* Copyright (c) Oleg O. Chukaev, 2007-2010  <oleg.chukaev@mail.ru>
* Based on AVRPP - AVR Parallel Programming Controller  R0.39 (C)ChaN, 2007
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <malloc.h>

#ifndef LINUXXX
# include <dir.h>
#endif

#include "defines.h"
#include "avr.h"
#include "common.h"
#include "debug.h"


extern char my_path[MAXPATH];


/* Search and Open configuration file */
FILE *open_cfgfile (char *filename) {

	FILE *fp;
	char tmp[MAXPATH], *p;
	char errstr[] = "open_cfgfile(): `%s': %s\n";

	fp = fopen (filename, "rt");
	debug_printf (2, errstr, filename, fp ? "Ok" : strerror (errno));
	if (fp)
		return (fp);

	strncpy (tmp, my_path, MAXPATH);
	strcat (tmp, filename);
	tmp[MAXPATH-1] = '\0';

	fp = fopen (tmp, "rt");
	debug_printf (2, errstr, tmp, fp ? "Ok" : strerror (errno));
	if (fp)
		return (fp);

#ifdef LINUXXX
	p = getenv ("HOME");
	if (p == NULL)
		return (NULL);
	strncpy (tmp, p, MAXPATH);
	strcat (tmp, "/.avrc/");
	strcat (tmp, filename);
	tmp[MAXPATH-1] = '\0';

	fp = fopen (tmp, "rt");
	debug_printf (2, errstr, tmp, fp ? "Ok" : strerror (errno));
	if (fp)
		return (fp);
#endif

	return (NULL);
}


/* fill device list */
/* , ⠬ ,  ࠡ⠥... */

int fill_dev_list (void) {

	int i, n, k = 0, ft, flw, eew, pd;
	FILE *f;
	char *z;
	char line[200], tmp[30];
	word m0, m1, m2, f0, f1, f2, ld, fu, ca, sc;
	dword fs, es;
	word fp, ep, ew, fw;
	unsigned int s0, s1, s2;

	for (i = 0; i < MAX_DEVICES; i++)
		dev_list[i] = NULL;

	f = open_cfgfile (CHIPSFILE);
	if (f == NULL) {
		/* !!! FIXME !!! */
		debug_printf (1, "fill_dev_list(): can't find `" CHIPSFILE "'\n");
		return (1);
	} else {
		n = 0;
		while (fgets (line, sizeof (line), f)) {
			z = strchr (line, '\r');
			if (z) *z = '\n';
			k++;
			if (line[0] == ';' || line[0] == '\n')
				continue;
			if ((dev_list[n] = malloc (sizeof(DEVPROP))) == NULL) {
				/* !!! FIXME !!! */
				debug_printf (2, "fill_dev_list(): can't allocate %d bytes for item %d\n",
						sizeof(DEVPROP), n);
				return (1);
			}
			i = sscanf (line,
	"%19s %x %x %x %lu %d %lu %d %d %d %x %d %d %x %x %x %x %x %x %d %d %d %x %d",
	tmp,  &s0, &s1, &s2, &fs, &fp, &es, &ep, &ew, &fw, &ld, &fu,
	&ca, &m0, &m1, &m2, &f0, &f1, &f2, &ft, &flw, &eew, &pd, &sc);
			if (i != 24) {
				/* !!! FIXME !!! */
				debug_printf (2, "fill_dev_list(): `" CHIPSFILE "': error on line %d: %d fields read\n", k, i);
				return (1);
			}
			strncpy (dev_list[n]->name, tmp, 19);
			dev_list[n]->sign[0] = (byte)s0;
			dev_list[n]->sign[1] = (byte)s1;
			dev_list[n]->sign[2] = (byte)s2;
			dev_list[n]->flash_size = fs;
			dev_list[n]->flash_page = fp;
			dev_list[n]->eeprom_size = es;
			dev_list[n]->eeprom_page = ep;
			dev_list[n]->erase_wait = ew;
			dev_list[n]->fuse_wait = fw;
			dev_list[n]->lock_data = ld;
			dev_list[n]->fuses = fu;
			dev_list[n]->cals = ca;
			dev_list[n]->fuse_mask[0] = (byte)m0;
			dev_list[n]->fuse_mask[1] = (byte)m1;
			dev_list[n]->fuse_mask[2] = (byte)m2;
			dev_list[n]->fuse_default[0] = (byte)f0;
			dev_list[n]->fuse_default[1] = (byte)f1;
			dev_list[n]->fuse_default[2] = (byte)f2;
			dev_list[n]->fuse_type = ft;
			dev_list[n]->flash_wait = flw;
			dev_list[n]->eeprom_wait = eew;
			dev_list[n]->poll_data = pd;
			dev_list[n]->socket = sc;
			n++;
		}
	}
	if (f != NULL)
		fclose (f);
	return (0);
}


/* kill device list */

void kill_dev_list (void) {

	int i;

	for (i = 0; dev_list[i]; i++) {
		free (dev_list[i]);
	}
}


/****************************************************************************
*  ᨢ ᠭ fuse-⮢.
* ࠬ:
*	---
* 頥:
*	0 --  Ok
*	1 -- -  ⠪...
****************************************************************************/

int fill_fuse_list (void) {

	int i, j, k, linenum = 0;
	/* !!! FIXME (size) !!! */
	char line[200], tmpstr[200], *cp, *z;
	FILE *f;

	for (i = 0; i < MAX_FUSES; i++) {
		fuse_names[i] = NULL;
		fuse_descr[i] = NULL;
	}

	f = open_cfgfile (FUSEFILE);
	if (f == NULL) {
		/* !!! FIXME !!! */
		debug_printf (1, "fill_fuse_list(): can't find `" FUSEFILE "'\n");
		return (1);
	}

	while (1) {
		linenum++;
		/* 諨   䠩,   ⠪   諨... */
		if (fgets (line, sizeof(line), f) == NULL) {
			fclose (f);
			/* !!! FIXME !!! */
			debug_printf (2, "fill_fuse_list(): can't find `%s' in `" FUSEFILE "'\n",
					device->name);
			return (1);
		}
		z = strchr (line, '\r');
		if (z) *z = '\n'; if (z && z[1] == '\n') z[1] = '\0';
		/* 饬 ^D.*{device->name}.* */
		if ((line[0] != 'D') || ((cp = strstr (line, device->name)) == NULL))
			continue;
		/*   ᮢ, 室 */
		if (strlen (cp) == strlen (device->name) + 1)
			break;
	}

	while (1) {
		linenum++;
		/* 諨   䠩,   ⠪   諨... */
		if (fgets (line, sizeof(line), f) == NULL) {
			fclose (f);
			/* !!! FIXME !!! */
			debug_printf (2, "fill_fuse_list(): can't find description for `%s'\n",
					device->name);
			return (1);
		}
		/* ᫥ ப "Mode: ?????" 稭 ᠭ ⮢ */
		if (strstr (line, "Mode:"))
			break;
	}

	while (1) {
		linenum++;
		/* 諨   䠩,   ⠪   諨... */
		if (fgets (line, sizeof(line), f) == NULL) {
			fclose (f);
			/* !!! FIXME !!! */
			debug_printf (2, "fill_fuse_list(): can't find description for `%s'\n",
					device->name);
			return (1);
		}
		/* ᠭ  稢 ⮩ ப */
		if (line[0] == '\n')
			break;

		/*  \n */
		cp = strchr (line, '\n');
		if (cp)
			*cp = '\0';

		switch (line[0]) {
		case 'L':
			i = 8 * 0; break;
		case 'H':
			i = 8 * 1; break;
		case 'E':
			i = 8 * 2; break;
		case 'K':
			i = 8 * 3; break;
		default:
			/* !!! FIXME !!! */
			debug_printf (2, "fill_fuse_list(): `" FUSEFILE "': invalid chars on line %d\n",
					linenum);
			return (1);
		}
		i += line[1] - '0';
		/* 㥬  fuse- */
		j = 4; k = 0;
		while (line[j] != '\0' && line[j] != ' ') {
			tmpstr[k++] = line[j++];
		}
		tmpstr[k] = '\0';
		fuse_names[i] = malloc (strlen (tmpstr) + 1);
		if (fuse_names[i] == NULL) {
			/* !!! FIXME !!! */
			debug_printf (2, "fill_fuse_list(): can't allocate %d bytes.\n", strlen (tmpstr) + 1);
			return (1);
		}
		strcpy (fuse_names[i], tmpstr);
		/* 㥬 ᠭ fuse- */
		fuse_descr[i] = malloc (strlen (line+j) + 1);
		if (fuse_descr[i] == NULL) {
			/* !!! FIXME !!! */
			debug_printf (2, "fill_fuse_list(): can't allocate %d bytes.\n", strlen (line+j));
			return (1);
		}
		if (line[j+1] != '\0') {
			strcpy (fuse_descr[i], line+j+1);
		} else {
			strcpy (fuse_descr[i], "\0");
		}
	}
	fclose (f);
	return (0);
}


/*  */
void kill_fuse_list (void) {

	int i;

	for (i = 0; i < MAX_FUSES; i++) {
		if (fuse_names[i] != NULL) {
			free (fuse_names[i]);
		}
		if (fuse_descr[i] != NULL) {
			free (fuse_descr[i]);
		}
	}
}

/** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* E0F
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **/

