/** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* BiDiPro/AVR v0.1.4                                              lin_drv.c
* Copyright (c) Oleg O. Chukaev, 2007-2010  <oleg.chukaev@mail.ru>
* Based on bidi12ar.asm - BiDiPro hardware driver by Eduard Panchenko,
* Alex Plusov / CnC Lab, Andrej Arnold
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **/


#include <stdio.h>
#include <stdlib.h>
#include <sys/io.h>
#include <unistd.h>


#define NELEMS(array)  (sizeof(array) / sizeof(array[0]))


enum {
	RGDC1 = 1,
	RGDC2 = 3,
	RGDP1 = 5,
	RGDIR = 6,
	RGDP2 = 7,
	RGDA1 = 9,
	RGDA2 = 11,
	RGDB1 = 13,
	RGDB2 = 15,
};


typedef unsigned char byte;
typedef unsigned short word;
typedef unsigned long dword;


int vcc_table[4] = { 45, 50, 55, 60 };
int vpp_table[32] = { 55, 65, 75, 85, 95, 105, 110, 115, 120, 125, 130, 135,
	140, 145, 150, 155, 160, 165, 170, 175, 180, 185, 190, 195, 200,
	205, 210, 220, 230, 240, 250, 260 };
int vid_table[2] = { 120, 140 };
word portaddr = 0;


static byte vpp_codes[32] = { 0x8f, 0x8e, 0x8d, 0x8c, 0x8b, 0x8a, 0x0f, 0x89,
	0x0e, 0x88, 0x0d, 0x87, 0x0c, 0x86, 0x0b, 0x85, 0x0a, 0x84, 0x09,
	0x83, 0x08, 0x82, 0x07, 0x81, 0x06, 0x80, 0x05, 0x04, 0x03, 0x02,
	0x01, 0x00 };
static byte rgd_tbl[16] = { 0 };


static word port (word val) {

	byte c, vl, vh;

	vh = (val >> 8) & 0xff;
	vl = val & 0xff;
	switch (vh & 0x0f) {
	case 0:			/* reset */
		port (0x859b);	/* ⠭ 1  In_In_In */
		port (0x8790);	/* ⠭ 2  In_Out_Out */
		port (0x83cf);	/* 殮  min */
		port (0x8600);	/* ⠭  "0" 22 */
		return (0);

	case 5:			/* wr p1 */
		rgd_tbl[RGDP1] = vl;
		rgd_tbl[RGDA1] = 0;
		rgd_tbl[RGDB1] = 0;
		rgd_tbl[RGDC1] = 0;
		break;

	case 7:			/* wr p2 */
		rgd_tbl[RGDP2] = vl;
		rgd_tbl[RGDA2] = 0;
		rgd_tbl[RGDB2] = 0;
		rgd_tbl[RGDC2] = 0;
		break;

	case 8:			/* rd state */
		c = inb (portaddr + 1);
		c = inb (portaddr + 1);
		return (c);

	case 9:			/* wr a1 */
	case 13:		/* wr b1 */
	case 1:			/* wr c1 */
	case 11:		/* wr a2 */
	case 15:		/* wr b2 */
	case 3:			/* wr c2 */
	case 6:			/* wr ir22 */
		if ((vh & 0x80) == 0 && rgd_tbl[vh] == vl)
			return (val);
		else
			rgd_tbl[vh] = vl;
		break;

	case 10:		/* rd a1 */
	case 14:		/* rd b1 */
	case 2:			/* rd c1 */
		outb (vh | 0x20, portaddr + 2);
		c = inb (portaddr);
		c = inb (portaddr);
		outb (vh & 0x0c, portaddr + 2);
		return (c);

	case 4:			/* nop */
	case 12:		/* nop */
		return (0);
	}
	outb (vl, portaddr);
	outb (vh & 0xdf, portaddr + 2);
	outb (vh & 0x0c, portaddr + 2);
	return (0);
}


static int findvlt (int *table, int size, int u) {

	int i;

	for (i = size - 1; i >= 0; i--) {
		if (table[i] <= u) {
			return (i);
		}
	}
	return (0);
}


static void oai9 (byte al, byte *c) {
	outb (al, portaddr + 2);
	*c <<= 1;
	if ((inb (portaddr + 1) & 8) == 0)
		*c += 1;
}


static byte testhard (void) {

	byte c, bh, al;

	c = inb (portaddr + 2);
	outb (c & 0xdf, portaddr);

	bh = 0x88;
	outb (bh, portaddr);
	c = 0;
	if (inb (portaddr) == bh)
		c = 1;
	outb (0x06, portaddr + 2);
	al = inb (portaddr + 1) & 0xf8;
	bh ^= al;
	c <<= 1;
	if ((bh & 0xb0) == 0)
		c += 1;
	c <<= 1;
	if ((al & 0x40) == 0)
		c += 1;
	oai9 (0x07, &c);
	oai9 (0x04, &c);
	oai9 (0x02, &c);
	oai9 (0x0e, &c);
	c <<= 1;
	if ((bh & 0x08) == 0)
		c += 1;
	outb (0x2c, portaddr + 2);
	return (~c);
}


int init (void) {

	byte c;

	c = inb (portaddr + 0x402);
	c = (c & 0x1f) | 0x20;
	outb (c, portaddr + 0x402);
	c = testhard ();
	if (c) {
		return (0x100 | c);
	} else {
		return (0);
	}
}


int voltage (word vcc, word vpp, word vid) {

	byte vb = 0;	/* /7/  p0 i0 c1 c0 p4 p3 p2 p1  /0/ */
	int z;

	vb |= findvlt (vcc_table, NELEMS(vcc_table), vcc) << 4;
	z = findvlt (vpp_table, NELEMS(vpp_table), vpp);
	vb |= vpp_codes[z];
	vb |= (findvlt (vid_table, NELEMS(vid_table), vid) ^ 1) << 6;
	port (0x8300 | vb);
	return (0);
}


word bdport (word cmd) {
	return (port (cmd));
}


void delay_us (word time) {

	word i;

	for (i = 0; i < time; i++) {
		inb (80);	/* Approx. 1 us */
	}
	/* usleep (time); */
}


int get_io_perm (void) {
	if (iopl (3)) {
		return (1);
	}
	return (0);
}


/** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* E0F
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **/

